/* BdfServer - Copyright (c) 2022-2023 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package fr.exemole.bdfserver.tools.exportation.transformation.defaulttemplates;

import fr.exemole.bdfserver.api.BdfServer;
import fr.exemole.bdfserver.api.managers.UiManager;
import fr.exemole.bdfserver.api.ui.CommentUi;
import fr.exemole.bdfserver.api.ui.DataUi;
import fr.exemole.bdfserver.api.ui.FieldUi;
import fr.exemole.bdfserver.api.ui.SubsetIncludeUi;
import fr.exemole.bdfserver.tools.exportation.transformation.BdfTransformationUtils;
import java.io.IOException;
import net.fichotheque.corpus.metadata.CorpusField;
import net.fichotheque.corpus.metadata.CorpusMetadata;
import net.fichotheque.corpus.metadata.FieldKey;
import net.mapeadores.util.xml.XsltWriter;


/**
 *
 * @author Vincent Calame
 */
public class AbstractFicheXsltWriter extends XsltWriter {

    protected final BdfServer bdfServer;
    protected final UiManager uiManager;
    protected final int extractVersion;
    protected final boolean uniqueLang;

    public AbstractFicheXsltWriter(BdfServer bdfServer, int extractVersion) {
        this.bdfServer = bdfServer;
        this.uiManager = bdfServer.getUiManager();
        this.extractVersion = extractVersion;
        this.uniqueLang = BdfTransformationUtils.isUniquelang(bdfServer);
    }

    public void appendCommentUi(CommentUi commentUi) throws IOException {
        if ((commentUi.getLocation() & CommentUi.TEMPLATE_BITVALUE) != 0) {
            addXslValueOf("$label[@name='" + commentUi.getName() + "']", true, true);
            addSeparator();
        }
    }

    public void appendFieldUi(FieldUi fieldUi, CorpusMetadata corpusMetadata) throws IOException {
        if (!acceptUi(fieldUi)) {
            return;
        }
        FieldKey fieldKey = fieldUi.getFieldKey();
        CorpusField corpusField = corpusMetadata.getCorpusField(fieldKey);
        String mode = "component-Inline";
        String select = null;
        switch (fieldKey.getCategory()) {
            case FieldKey.PROPRIETE_CATEGORY:
                select = "propriete[@name='" + fieldKey.getFieldName() + "']";
                break;
            case FieldKey.SECTION_CATEGORY:
                select = "section[@name='" + fieldKey.getFieldName() + "']";
                mode = "component-Section";
                break;
            case FieldKey.INFORMATION_CATEGORY:
                select = "information[@name='" + fieldKey.getFieldName() + "']";
                mode = DefaultTemplateUtils.getMode(corpusField);
                break;
            case FieldKey.SPECIAL_CATEGORY:
                switch (fieldKey.getKeyString()) {
                    case FieldKey.SPECIAL_LANG:
                        select = "lang";
                        break;
                    case FieldKey.SPECIAL_REDACTEURS:
                        select = "redacteurs";
                        break;
                }
                break;
        }
        if (select != null) {
            apply(select, mode, fieldUi.getFieldString());
        }
    }


    public void appendSubsetIncludeUi(SubsetIncludeUi includeUi) throws IOException {
        String select = getTagName(includeUi.getCategory()) + "[@name='" + includeUi.getName() + "']";
        String mode = DefaultTemplateUtils.getMode(includeUi);
        apply(select, mode, includeUi.getName());
    }

    public void appendDataUi(DataUi dataUi) throws IOException {
        apply("data[@name='" + dataUi.getDataName() + "']", "component-Listitem", dataUi.getName());
    }

    public void appendChrono() throws IOException {
        openXslApplyTemplates(".", "component-Chrono", true);
        addXslWithParam("param_CreationLabel", "$label[@name='date_creation']", false);
        addXslWithParam("param_ModificationLabel", "$label[@name='date_modification']", false);
        closeXslApplyTemplates(true);
        addSeparator();
    }

    public void apply(String match, String mode) throws IOException {
        appendIndent();
        addXslApplyTemplates(match, mode);
        addSeparator();
    }

    public void apply(String select, String mode, String labelName) throws IOException {
        openXslApplyTemplates(select, mode, true);
        addXslWithParam("param_Label", "$label[@name='" + labelName + "']", false);
        closeXslApplyTemplates(true);
        addSeparator();
    }

    public String getTagName(short subsetCategory) {
        return DefaultTemplateUtils.getTagName(subsetCategory, extractVersion);
    }

    public boolean acceptUi(FieldUi fieldUi) {
        switch (fieldUi.getFieldString()) {
            case FieldKey.SPECIAL_TITRE:
            case FieldKey.SPECIAL_SOUSTITRE:
                return false;
            case FieldKey.SPECIAL_LANG:
                return (!uniqueLang);
            default:
                return true;
        }
    }

}
