/* FichothequeLib_API - Copyright (c) 2006-2022 Vincent Calame - Exemole
 * Logiciel libre donné sous triple licence :
 * 1) selon les termes de la CeCILL V2
 * 2) selon les termes de l’EUPL V.1.1
 * 3) selon les termes de la GNU GPLv3
 * Voir le fichier licences.txt
 */


package net.fichotheque.utils;

import net.fichotheque.corpus.metadata.CorpusField;
import net.fichotheque.corpus.metadata.FieldKey;
import net.fichotheque.corpus.metadata.FieldOptionException;
import net.fichotheque.corpus.metadata.FieldOptionsConstants;


/**
 * Classe utilitaire de traitement des options.
 *
 * @author Vincent Calame
 */
public final class FieldOptionsUtils {


    private FieldOptionsUtils() {
    }

    /**
     * Retourne la valeur de l'option INFORMATIONDISPLAY_OPTION qui doit être
     * utilisée par défaut (il s'agit de INLINE_DISPLAY).
     */
    public static short getDefaultInformationDisplay() {
        return FieldOptionsConstants.INLINE_DISPLAY;
    }


    /**
     *
     * @throws IllegalArgumentException
     */
    public static String informationDisplayToString(short informationDisplay) {
        switch (informationDisplay) {
            case FieldOptionsConstants.INLINE_DISPLAY:
                return "inline";
            case FieldOptionsConstants.BLOCK_DISPLAY:
                return "block";
            default:
                throw new IllegalArgumentException("wrong value");
        }
    }

    /**
     *
     * @throws IllegalArgumentException
     */
    public static short informationDisplayToShort(String informationDisplay) {
        switch (informationDisplay) {
            case "inline":
                return FieldOptionsConstants.INLINE_DISPLAY;
            case "block":
                return FieldOptionsConstants.BLOCK_DISPLAY;
            default:
                throw new IllegalArgumentException("Unknown value: " + informationDisplay);
        }
    }

    public static short getDefaultLangScope(CorpusField corpusField) {
        return getDefaultLangScope(corpusField.getFieldKey());
    }

    public static short getDefaultLangScope(FieldKey fieldKey) {
        switch (fieldKey.getKeyString()) {
            case FieldKey.SPECIAL_LANG:
                return FieldOptionsConstants.CONFIG_SCOPE;
            default:
                return FieldOptionsConstants.ALL_SCOPE;
        }
    }

    /**
     *
     * @throws IllegalArgumentException
     */
    public static String langScopeToString(short langScope) {
        switch (langScope) {
            case FieldOptionsConstants.ALL_SCOPE:
                return "all";
            case FieldOptionsConstants.CONFIG_SCOPE:
                return "config";
            case FieldOptionsConstants.LIST_SCOPE:
                return "list";
            default:
                throw new IllegalArgumentException("Unknown value: " + langScope);
        }
    }

    /**
     *
     * @throws IllegalArgumentException
     */
    public static short langScopeToShort(String langScope) {
        switch (langScope) {
            case "all":
                return FieldOptionsConstants.ALL_SCOPE;
            case "config":
                return FieldOptionsConstants.CONFIG_SCOPE;
            case "list":
                return FieldOptionsConstants.LIST_SCOPE;
            default:
                throw new IllegalArgumentException("Unknown value: " + langScope);
        }
    }

    public static boolean haveSubfield(int ficheItemType) {
        switch (ficheItemType) {
            case CorpusField.PERSONNE_FIELD:
            case CorpusField.GEOPOINT_FIELD:
            case CorpusField.IMAGE_FIELD:
                return true;
            default:
                return false;
        }
    }

    public static boolean testFieldOption(String optionName, CorpusField corpusField) {
        try {
            checkFieldOption(optionName, corpusField.getFieldKey(), corpusField.getFicheItemType());
            return true;
        } catch (FieldOptionException foe) {
            return false;
        }
    }

    public static boolean testFieldOption(String optionName, FieldKey fieldKey, int ficheItemType) {
        try {
            checkFieldOption(optionName, fieldKey, ficheItemType);
            return true;
        } catch (FieldOptionException foe) {
            return false;
        }
    }

    public static void checkFieldOption(String optionName, FieldKey fieldKey, int ficheItemType) throws FieldOptionException {
        switch (optionName) {
            case FieldOptionsConstants.INFORMATIONDISPLAY_OPTION:
                if (!fieldKey.isInformation()) {
                    throw new FieldOptionException("informationDisplay Option applies only to Information Field");
                }
                break;
            case FieldOptionsConstants.DEFAULTSPHEREKEY_OPTION:
                if (ficheItemType != CorpusField.PERSONNE_FIELD) {
                    throw new FieldOptionException("ficheItemType must be PERSONNE_FIELD");
                }
                break;
            case FieldOptionsConstants.SUBFIELDDISPLAY_OPTION:
                if (!fieldKey.isPropriete()) {
                    throw new FieldOptionException("subfieldDisplay Option applies only to Propriete Field");
                }
                if (!haveSubfield(ficheItemType)) {
                    throw new FieldOptionException("wrong ficheItemType = " + ficheItemType);
                }
                break;
            case FieldOptionsConstants.CURRENCYARRAY_OPTION:
                if (ficheItemType != CorpusField.MONTANT_FIELD) {
                    throw new FieldOptionException("ficheItemType must be MONTANT_FIELD");
                }
                break;
            case FieldOptionsConstants.ADDRESSFIELDARRAY_OPTION:
                if (!fieldKey.isPropriete()) {
                    throw new FieldOptionException("addressFields Option applies only to Propriete Field");
                }
                if (ficheItemType != CorpusField.GEOPOINT_FIELD) {
                    throw new FieldOptionException("ficheItemType must be GEOPOINT_FIELD");
                }
                break;
            case FieldOptionsConstants.BASEURL_OPTION:
                if ((ficheItemType != CorpusField.IMAGE_FIELD) && (ficheItemType != CorpusField.LINK_FIELD)) {
                    throw new FieldOptionException("ficheItemType must be IMAGE_FIELD or lINK_FIELD");
                }
                break;
            case FieldOptionsConstants.LANGSCOPE_OPTION:
                if (!isLangField(fieldKey, ficheItemType)) {
                    throw new FieldOptionException("langScope applies only to FieldKey.LANG and fields with ficheItemType = LANGUE_FIELD");
                }
                break;
            case FieldOptionsConstants.LANGARRAY_OPTION:
                if ((!isLangField(fieldKey, ficheItemType)) && (ficheItemType != CorpusField.PARA_FIELD)) {
                    throw new FieldOptionException("langArray applies only to FieldKey.LANG and fields with ficheItemType = PARA_FIELD or LANGUE_FIELD");
                }
                break;
            case FieldOptionsConstants.GEOLOCALISATIONFIELD_OPTION:
                if ((!fieldKey.isPropriete()) || (ficheItemType != CorpusField.GEOPOINT_FIELD)) {
                    throw new FieldOptionException("geolocalisationField applies only to Propriete Field with ficheItemType = GEOPOINT_FIELD");
                }
                break;
        }
    }

    private static boolean isLangField(FieldKey fieldKey, int ficheItemType) {
        if (ficheItemType == CorpusField.LANGUE_FIELD) {
            return true;
        }
        switch (fieldKey.getKeyString()) {
            case FieldKey.SPECIAL_LANG:
                return true;
            default:
                return false;
        }
    }

}
